--[[
	OPTIONS
]]

--If you want to unload global variables and functions from scripts
local unload_globals = true

--Scripts path
local path = "sys/lua/"

--Colors
local errColor = "255000000"
local okColor = "000255000"
local infoColor = "200100200"




--Utilites
dofile(path.."script_manager/util.lua")


--Loaded scripts and their hooks/funcs
local loaded = {}

--Loaded hooks (temp)
local curHooks = {}

--Full script "names" for resetting (temp)
local fullSave = {}

--Included directories
local includes = {}

--Environments
--Epic (I guess) thing which allows us to unload everything global
local envs
if(unload_globals) then
	envs = {}
end

--Check if the script is loaded
function scriptLoaded(s_name)
	if(s_name ~= nil) then
		for name,_ in pairs(loaded) do
			if(name == s_name) then
				return true
			end
		end
	end
	return false
end

--Directory inclusion
function include(directory)
	--Just in case, because you can use bad auto-loaders
	if((type(directory) ~= "string") or (#directory == 0)) then
		print(errColor.."Failed to include a directory")
		return
	end
	directory = trim(directory,"/")
	if(#includes > 0) then
		for c = 1,#includes do
			if(includes[c] == directory) then
				print(errColor.."Failed to include a directory : directoy "..directory.." is already included!")
				return
			end
		end
	end
	table.insert(includes, directory)
end

--Load scripts after the reset
if(fileExists("tempScripts.txt")) then
	for line in io.lines("tempScripts.txt") do
		loadscript(line)
	end
	os.remove("tempScripts.txt")
end


--[[
		Help
]]
--Contents
local cmdhelp = {}

--Function to add some help
function addCmdHelp(cmd,...)
	--You cann append to existing "help"
	if(cmdhelp[cmd] == nil) then
		cmdhelp[cmd] = {}
	end
	for c = 1, #arg do
		table.insert(cmdhelp[cmd],arg[c])
	end
end

addCmdHelp("load","load \"scriptname1\" \"scriptname2\"...","Load the script(s)")

addCmdHelp("unload","unload \"scriptname1\" \"scriptname2\"...","Unload the script(s)")
if(unload_globals) then
	addCmdHelp("unload","WARNING!","Removes all global variables and functions")
end

addCmdHelp("reset","reset","Resets the game, reloads the scripts","WARNING!","It saves only scripts, this command doesn't save positions, money etc.")

addCmdHelp("show","show \"scriptname\"","Shows script hooks and functions","scriptname = blank : shows the loaded scripts","scriptname = full : shows the loaded scripts with extensions and directories")

addCmdHelp("help","help \"cmd\"","Shows some info about cmd","cmd = blank : show all the commands")

addCmdHelp("include","include \"dir1\" \"dir2\"...","Include directories for loading")


--"Addhook" copy
local addhook_default = addhook

--Slightly modified "addhook"
function addhook_loading(hook,func,priority)
	if(tonumber(priority) == nil) then
		priority = 0
	end
	addhook_default(hook,func,priority)
	if(curHooks[hook] == nil) then
		curHooks[hook] = {}
	end
	table.insert(curHooks[hook],func)
end


--Our "loadscript" function
function loadscript(filename)
	--Oops, wrong parameter type!
	--Just in case, because you can use bad auto-loaders
	if(type(filename)~="string") then
		print(errColor.."Wrong parameter type")
		return
	end
	--Extension
	local ext = ""
	if(filename:sub(#filename - 3) ~= ".lua") then
		ext = ".lua"
	end
	--/sys/lua/ thingie is optional
	if(filename:sub(1,#path) == path) then
		filename = filename:sub(#path + 1)
	end

	--Do we nned to load it?
	local toLoad = false
	--Inclusion
	local include = ""
	--Using weak keys for table messed everything up
	--So I'll use the primitive way
	if(fileExists(path..filename..ext)) then
		toLoad = true
	end
	--It's not in the main folder...
	if(not toLoad) then
		--...So we have to search the included ones
		for c = 1, #includes do
			if(not toLoad) then
				include = includes[c].."/"
				if(fileExists(path..include..filename..ext)) then
					toLoad = true
				end
			end
		end
	end

	--Load the script
	if(toLoad) then
		--Get the file
		local file = include..filename..ext
		--Find out if our script is in directory
		--You don't want to remember its directory every time, do you?
		local dir = file:reverse():find("/",1,true)
		dir = dir or (#file + 1)
		--If dir == nil then dir will be = 1
		dir = #file - dir + 2
		local scriptname = file:sub(dir,#file-4)
		--Get a fresh copy of _G
		local envCopy = {}
		if(unload_globals) then
			--deepCopy will cause overflow
			--Thus, the manager won't unload overloaded functions
			for key,value in pairs(_G) do
				envCopy[key] = value
			end
			envs[scriptname] = {}
		end
		--Prepare the table for hooks 'n functions
		loaded[scriptname] = {}

		--Change the "addhook"
		addhook = addhook_loading
		--"Do" the file!
		dofile(path..file)
		--Then swap our addhook back
		addhook = addhook_default

		--Get the difference between the current and modified _G
		if(unload_globals) then
			for key, value in pairs(_G) do
				if((envCopy[key] == nil) or (envCopy[key] ~= value)) then
					envs[scriptname][key] = value
				end
			end
		end

		--"addhook_loading" adds all the hooks to the "curHooks" table...
		loaded[scriptname] = deepCopy(curHooks)
		--..so we need to clear it after using
		curHooks = {}

		--Save the name for future resets
		fullSave[scriptname] = file

		--Hooray, script loaded!
		print(okColor..filename..ext.." loaded!")
	else
		--You can't load scripts if you don't have 'em
		print(errColor.."Failed to load "..filename..ext..":No such file/wrong directory")
	end
end

--"Parse" hook
addhook("parse","scriptControl")
function scriptControl(text)
	local parsed = parseLine(text)

	--Unload script
	if(parsed[1] == "unload") then
		if(#parsed > 1) then
			for c = 2, #parsed do
				local ext = ""
				if(parsed[c]:sub(#parsed[c] - 3) ~= ".lua") then
					ext = ".lua"
				end
				--We can't unload scripts we haven't loaded, right?
				if(scriptLoaded(parsed[c])) then
					--Free all hooks
					for hook,hookTab in pairs(loaded[parsed[c]]) do
						for c = 1, #hookTab do
							freehook(hook,hookTab[c])
						end
					end
					--"Forget" about loading it after resets
					fullSave[parsed[c]] = nil
					--Free its hook table
					loaded[parsed[c]] = nil
					--Unload its environment, if needed
					if(unload_globals) then
						--We just have to erease the difference in its environment
						--and the default _G...
						for key,_ in pairs(envs[parsed[c]]) do
							_G[key] = nil
						end
						--...And then "forget" it too
						envs[parsed[c]] = nil
					end
					print(infoColor.."Unloaded "..parsed[c]..ext)
				else
					--ERROR
					print(errColor..parsed[c]..ext.." : No script loaded")
				end
			end
			--We have a lot of garbage now
			collectgarbage()
		else
			--ERROR
			print(errColor.."No script specified!")
			print(errColor.."Usage:")
			print(errColor.."-"..cmdhelp[parsed[1]])
		end
		return 1

	--Load script
	elseif(parsed[1] == "load") then
		if(#parsed > 1) then
			for c = 2, #parsed do
				loadscript(parsed[c])
			end
		else
			--ERROR
			print(errColor.."No script specified!")
			print(errColor.."Usage:")
			print(errColor.."-"..cmdhelp[parsed[1]])
		end
		return 1

	--Show scripts
	elseif(parsed[1] == "show") then
		if(scriptLoaded(parsed[2])) then
			print(infoColor..parsed[2]..":")
			for hook,func in pairs(loaded[parsed[2]]) do
				print(infoColor.."-Hook : "..hook)
				for c = 1,#func do
					print(infoColor.."-Function : "..func[c])
				end
			end
		else
			local lookup = loaded
			if(parsed[2]) then
				--Kinda hacky way to implement this
				if(parsed[2] == "full") then
					lookup = {}
					--But it does its job
					for key, value in pairs(fullSave) do
						lookup[value] = ""
					end
				else
					print(errColor.."No script found : "..parsed[2])
				end
			end
			print(infoColor.."Loaded scripts:")
			for script,_ in pairs(lookup) do
				print(infoColor.."-"..script)
			end
		end
		return 1

	--Reset everything
	elseif(parsed[1] == "reset") then
		--Save scripts
		local temp = io.open("tempScripts.txt","w")
		for script,_ in pairs(loaded) do
			temp:write(fullSave[script])
		end
		temp:close()

		print(infoColor.."Game reset!")
		parse("restart")
		return 1

	--Inclusion
	elseif(parsed[1] == "include") then
		if(#parsed > 1) then
			for c = 2, #parsed do
				include(parsed[c])
				print(infoColor.."Included "..parsed[c])
			end
		else
			--ERROR
			print(errColor.."No directories specified!")
			print(errColor.."Usage:")
			print(errColor.."-"..cmdhelp[parsed[1]])
		end
		return 1

	--Help
	elseif(parsed[1] == "help") then
		if(#parsed == 2) then
			if(cmdhelp[parsed[2]] ~= nil) then
				print(infoColor.."Info:")
				for c = 1, #cmdhelp[parsed[2]] do
					print(infoColor.."-"..cmdhelp[parsed[2]][c])
				end
			else
				--ERROR
				print(errColor..parsed[2].." : No such command!")
			end
		elseif(#parsed == 1) then
			print(infoColor.."Available commands:")
			for key,_ in pairs(cmdhelp) do
				print(infoColor.."-"..key)
			end
		else
			--ERROR
			print(errColor.."Usage:")
			print(errColor.."-"..cmdhelp[parsed[1]])
		end
		return 1

	--Unknown command
	else
		return 0
	end
end
